/*********************************************************************************************
 *** Fischertechnik Universal Interface 30520 als Porterweiterung am Atmel AVR
 *********************************************************************************************
 *** INKLUSIVE PWM MOTORSTEUERUNG ************************************************************
 *********************************************************************************************
 +++ Es sind zwei Varianten verfgbar, bitte nur EINE *.c Datei dazu verlinken +++
 1) Variante mit Nutzung eines einzigen Interrupts (Timer 2 Overflow) -> FTUI_PWM.c
 2) Variante mit Nutzung von bis zu 8 OCR Interrupts im ATMEGA 2560 -> FTUI_PWM_OCR.c
 *********************************************************************************************
 
 Es wird keine Haftung bernommen! Die Benutzung dieser Routine geschieht
 auf eigene Gefahr!
 
 Autor:			Dirk Uffmann, mailto: uffmann@arcor.de
 
 Datum:			27.11.2015
 
 Umgebung:		Atmel Studio 6.2, X-Loader, Arduino Bootloader, ATMEGA2560@16MHZ

 Beschreibung:	Programm zur Nutzung des parallelen Fischertechnik interface 30520 (FTUI)
				an einem Atmel AVR Mikrocontroller.
				
				Folgende Funktionalitt wird untersttzt:
				- ein oder zwei interfaces (im Master- und Slave-Betrieb) 
				- 8 (bzw. 16) digitale Eingnge
				- 4 (bzw. 8) Motoren
				
				Die analogen Eingnge Ex, Ey werden direkt mit Analog-Inputs am AVR
				verbunden. 
				
				Der AVR wird mittels 5 digitalen I/O PORTS an das FTUI angeschlossen.
				Die Pinzuordnung wird hier in FTUI.h festgelegt. Es wird auerdem 
				festgelegt, ob ein oder zwei interfaces angeschlossen sind.

				Das FTUI wird wie folgt mit dem Arduino MEGA2560 Board verbunden:

				MEGA2560-(PORT K)-----FTUI-(20-pol. Pfostenstecker intern)
				-------------------------------------------------------------------
				PK1-------------LoadIn-(20P Pin  7)
				PK2------------LoadOut-(20P Pin  8)
				PK3--------------Clock-(20P Pin  9)
				PK4------------DataOut-(20P Pin 10)
				PK5-----------------EY-(20P Pin 13)
				PK6-----------------EX-(20P Pin 14)
				PK7------------DataIn*-(20P Pin 17)
				GND----------------GND-(20P Pin 19)

				* Achtung: Eigentlich liegt DataIn auf Pin 18. Der Ausgang des Eingangs-
				schieberegisters wird auf dem Weg zu Pin 18 ber ein Oder-Gatter mit dem 
				Ausgangssignal der NE556 Timer der beiden Analogeingnge verknpft und
				dann ber einen Treibertransistor BC548 invertiert.
				Da wir die Timersignale nicht brauchen und diese sogar 
				stren, wenn die Triggersignale dafr nicht auf high gelegt werden und die
				analogen Eingnge nicht durchgeschleift werden auf den Timer, verwenden
				wir lieber Pin17 als direkten Ausgang des Schieberegisters. Hier hat das
				Ausgangssignal auch noch eine geringere Verzgerungszeit zur positiven 
				Clockflanke und daher knnen wir dann mit einer hheren bertragungsrate
				arbeiten.
				 
				Die Pinreihenfolge ist hier so zum AVR PORT zugeordnet, 
				dass der interne Pfostenstecker im FT-Interface gut angeschlossen 
				werden kann (passende Reihenfolge auf dem Flachbandkabel). 

				Bevor das FTUI genutzt werden kann, muss FTUI_Init() aufgerufen werden.
				Anschlieend kann ber die Funktion
				
				M(Motor_Nr, Aus/An & Drehrichtung, pwm)

				Zulssige Wertebereiche der drei ganzzahligen Parameter:
				Motor_Nr: 1-8
				state bzw. Aus/An & Drehrichtung: 0-3
				pwm: 0-100
					
				die Variable motorcontrol zur Steuerung der Motoren  kodiert werden.
				 
				Um die Einstellungen am FT-Interface wirksam werden zu lassen, wird 
				FTUI_update() alle 12,8 ms aus der Interrupt Service Routine zum
				Timer 1 Overflow aufgerufen.

				Mit der Funktion E(Eingangs_Nr) kann der Status eines Schalters aus der 
				eingelesenen Variable inputword dekodiert werden. 
				
				Die interne Funktion FTUI_update() muss mindestens alle 300ms aufgerufen werden, 
				da sonst das FTUI davon ausgeht, dass der Steuercomputer abgestrzt ist
				(dies passiert automatisch durch den Timer Interrupt mindestens alle 12,8 ms).

********************************************************************************************/
#ifndef _FTUI_PWM_H
 #define _FTUI_PWM_H

#include <stdio.h>
#include <inttypes.h>
#include <avr/interrupt.h>
#include <avr/io.h>

/*******************************************************************************
	Port, an dem das parallele Fischertechnik Interface 30520 am AVR
	angeschlossen ist, ggf. hier anpassen:
*******************************************************************************/
#define FTUI_Port_DDR		DDRK
#define FTUI_Port_Write		PORTK
#define FTUI_Port_Read		PINK

/*******************************************************************************
	Pin-Belegung am Port
	ggf. hier anpassen:
*******************************************************************************/

#define FTUI_LoadIn 		1
#define FTUI_LoadOut		2
#define FTUI_Clock 			3
#define FTUI_DataOut 		4
#define FTUI_DataIn 		7

/*******************************************************************************
	Definiere, ob am Erweiterungsport des ersten Master-Interfaces  
	noch ein zweites Slave-Interface angeschlossen ist. 

	Falls nein:	slave=0
	Falls ja:	slave=1

	ggf. hier anpassen:
*******************************************************************************/
#define slave				0

/*******************************************************************************
	Motoren und andere zweiphasig angeschlossenen Verbraucher werden im 
	Ruhezustand auf V+ geschaltet, um den Stromverbrauch der TLE4201 
	zu senken und deren	Erwrmung zu mindern. 
	Definiere die Beschaltung der Ausgnge, hier am Beispiel M1:
	- #define M1 1  -> im Ruhezustand sind beide Ausgnge auf +V

	Verbraucher, die einphasig gegen Masse geschaltet sind, werden im 
	Ruhezustand auf GND gelegt. Hier am Beispiel der beiden Ausgnge an M1:
	- #define M1 0 -> im Ruhezustand sind beide Ausgnge auf GND
*******************************************************************************/
#define M1		1	//Ruhezustand ist, wenn beide Ausgnge auf +V liegen (state = 3)
#define M2		1
#define M3		1
#define M4		0	//Ruhezustand ist, wenn beide Ausgnge auf GND liegen (state = 0)
#define M5		1
#define M6		1
#define M7		1
#define M8		1
//ab hier nichts mehr ndern
#define output1	(((M4*3)<<6) + ((M3*3)<<4) + ((M2*3)<<2) + (M1*3))
#define output2	(((M8*3)<<6) + ((M7*3)<<4) + ((M6*3)<<2) + (M5*3))
#define output ((output2<<8) + output1)

/*******************************************************************************
	Initialisierung, bitte vor der Nutzung von FTUI einmal aufrufen
*******************************************************************************/
extern void FTUI_Init (void);

/*******************************************************************************
	M() kodiert bzw. modifiziert das motorword fr den jeweiligen Motor
	M (motor_number,state)
	motor_number 1...4 bzw. 1...8 (mit zweitem Slave-Interface)
	state=0 => Motor aus / Minus an beiden Motorausgngen
	state=1 => Motor an / Plus am Motorausgang links (z.B orange bei M1)
	state=2 => Motor an / Plus am Motorausgang rechts (z.B. gelb bei M1)
	state=3 => Motor aus / Plus an beiden Motor-Kabeln
*******************************************************************************/
extern void M (uint8_t motor_number, uint8_t state, uint8_t pwm); 

/*******************************************************************************
	E () dekodiert den Zustand der Eingnge aus dem inputword
	E(input_number)
	input_number 1...8 bzw. 1...16
*******************************************************************************/
extern uint8_t E (uint8_t input_number); 

#endif //_FTUI_PWM_H
