/***************************************************************************************
	Fischertechnik Universal Interface 30520 als Porterweiterung am Atmel AVR.

*** INKLUSIVE PWM MOTORSTEUERUNG ***
*** Variante mit Nutzung von bis zu 8 OCR Interrupts im MEGA 2560 ***

	Weitere Hinweise siehe FTUI_PWM.h

	Es wird keine Haftung bernommen! Die Benutzung dieser Routine geschieht
	auf eigene Gefahr!

	Autor:	Dirk Uffmann, mailto: uffmann@arcor.de

	Datum:	27.11.2015

***************************************************************************************/

#include "FTUI_PWM.h"

#if slave
	volatile uint16_t motorword;
	volatile uint16_t motorcontrol;
	volatile uint16_t inputword;
#else
	volatile uint8_t motorword;
	volatile uint8_t motorcontrol;
	volatile uint8_t inputword;
#endif

/******************************************************************************
	FTUI_update bertrgt das motorword und liest parallel das inputword ein.
	Diese Funktion muss mindestens alle 300 ms aufgerufen werden, um den 
	Watchdog im FT-Interface zurckzusetzen (durch das Clocksignal). Das 
	passiert hier automatisch ber den Timer1-Overflow-Interrupt alle 12,8 ms.
	Bei einem Systemtakt von 16 MHz braucht die Routine mit einem Interface
	nur ca. 12 s, mit zwei Interfaces sind es ca. 32 s.
******************************************************************************/
inline void FTUI_update (void)
{
#if slave
	uint16_t mask_motor=(1<<15);
	uint16_t mask_input=(1<<7);
#else
	uint8_t mask_motor=(1<<7);
	uint8_t mask_input=(1<<7);
#endif
//	Clock low-Phase einleiten
	FTUI_Port_Write &= ~(1<<FTUI_Clock); 

//	Zunchst werden alle Eingnge als 0 angenommen.
	inputword = 0;

//	CMOS4014: Nun werden die Eingangszustnde in das Schieberegister bernommen.
//	Dazu mu LoadIn gesetzt werden und danach eine positive Clockflanke erzeugt werden.
	FTUI_Port_Write |= (1<<FTUI_LoadIn);
	FTUI_Port_Write |= (1<<FTUI_Clock); //Clock high-Phase einleiten
//	mit dieser positiven Clockflanke steht das erste Bit auf der
//	DataIn-Leitung an, jedoch mit einer Verzgerung. Daher wird
//	es erst in der nchsten Clock-Low-Phase eingelesen

//	CMOS4014: Dann werden die Eingnge wieder vom Schieberegister entkoppelt
	FTUI_Port_Write &= ~(1<<FTUI_LoadIn);

//	Dann werden die Bits 7 bis 0 (ein FT-Interface) bzw. 15 bis 0 (zwei
//	FT-Interfaces) gesendet (motorword) und parallel die Bits 7 bis 0 und
//	dann 15 bis 8 (zwei FT-Interfaces) eingelesen (inputword)
	while (mask_motor)
	{
//		Wenn das jeweilige Sende-Bit high sein soll, dann wird DataOut = 1
//		gesetzt, sonst DataOut = 0
		if(motorword & mask_motor) FTUI_Port_Write |= (1<<FTUI_DataOut);
		else FTUI_Port_Write &= ~(1<<FTUI_DataOut);
		FTUI_Port_Write &= ~(1<<FTUI_Clock); //Clock low-Phase einleiten
		mask_motor = (mask_motor>>1); // Schiebe das Maskierungsbit eins nach rechts

//		Nun wird das jeweilige Bit im Datenstream auf der Leitung DataIn
//		daraufhin getestet, ob der zugehrige Eingang "high" ist,
//		und dann wird inputword an der jeweilige Bitposition auf 1 korrigiert
		if (FTUI_Port_Read & (1<<FTUI_DataIn)) inputword |= mask_input;
		mask_input = (mask_input>>1); //Maskierungsbit nach rechts weiterschieben
#if slave
		if (mask_input==0) mask_input = (1<<15); //Maskierungsbit auf Position 15 setzen
#endif
		FTUI_Port_Write |= (1<<FTUI_Clock); //Clock high-Phase einleiten
	}

//	CMOS4094: Durchschalten der Daten aus dem Schieberegister ins Speicherregister
//	fr die Motorausgnge, dieses Strobe-Signal muss 200ns andauern (> 3 Takte 16MHz)
	FTUI_Port_Write |= (1<<FTUI_LoadOut);
	FTUI_Port_Write |= (1<<FTUI_LoadOut);
	FTUI_Port_Write |= (1<<FTUI_LoadOut);

//	wieder Entkoppeln von Speicherregister und Schieberegister
	FTUI_Port_Write &= ~(1<<FTUI_LoadOut);
}

ISR (TIMER1_OVF_vect)
{
//	set all outputs according to the variable motorcontrol
	motorword = motorcontrol;
	FTUI_update();
	TCNT3 = 0; //synchronize all relevant timers 
	TCNT4 = 0;
	TCNT5 = 0;
}

ISR (TIMER1_COMPA_vect)
{

#if M1
//	switch off motor 1 according to value in OCR1A
	motorword |= 0x03;//switch the relevant motor off
#else
	motorword &= ~0x03;//switch the relevant motor off
#endif
	FTUI_update();
}

ISR (TIMER1_COMPB_vect)
{
#if M2
//	switch off motor 2 according to value in OCR1B
	motorword |= 0x0C;//switch the relevant motor off
#else
	motorword &= ~0x0C;//switch the relevant motor off
#endif
	FTUI_update();
}

ISR (TIMER1_COMPC_vect)
{
#if M3
//	switch off motor 3 according to value in OCR1C
	motorword |= 0x30;//switch the relevant motor off
#else
	motorword &= ~0x30;//switch the relevant motor off
#endif
	FTUI_update();
}

ISR (TIMER3_COMPA_vect)
{
#if M4
//	switch off motor 4 according to value in OCR3A
	motorword |= 0xC0;//switch the relevant motor off
#else
	motorword &= ~0xC0;//switch the relevant motor off
#endif
	FTUI_update();
}

#if slave
ISR (TIMER3_COMPB_vect)
{
#if M5
	//	switch off motor 5 according to value in OCR3B
	motorword |= 0x0300;//switch the relevant motor off
#else
	motorword &= ~0x0300;//switch the relevant motor off
#endif
	FTUI_update();
}

ISR (TIMER3_COMPC_vect)
{
#if M6
	//	switch off motor 6 according to value in OCR3C
	motorword |= 0x0C00;//switch the relevant motor off
#else
	motorword &= ~0x0C00;//switch the relevant motor off
#endif
	FTUI_update();
}

ISR (TIMER4_COMPA_vect)
{
#if M7
	//	switch off motor 7 according to value in OCR4A
	motorword |= 0x3000;//switch the relevant motor off
#else
	motorword &= ~0x3000;//switch the relevant motor off
#endif
	FTUI_update();
}

ISR (TIMER4_COMPB_vect)
{
#if M8
	//	switch off motor 7 according to value in OCR4A
	motorword |= 0xC000;//switch the relevant motor off
#else
	motorword &= ~0xC000;//switch the relevant motor off
#endif
	FTUI_update();
}
#endif

/****************************************************************************
	FTUI Init Routine
****************************************************************************/
void FTUI_Init (void)
{
//	internen Pull-up Widerstand fr DataIn einschalten
	FTUI_Port_Write = (1<<FTUI_DataIn);
//	Ausgnge auf low setzen
	FTUI_Port_Write &= ~((1<<FTUI_LoadIn)|(1<<FTUI_DataOut)|(1<<FTUI_Clock)|(1<<FTUI_LoadOut));
//	Ausgnge definieren bzw. aktiv schalten
	FTUI_Port_DDR = (1<<FTUI_LoadOut)|(1<<FTUI_LoadIn)|(1<<FTUI_DataOut)|(1<<FTUI_Clock);
//	system clock 16 MHz is used (external quartz oscillator)
//	setup timers and compare matches
//	fast PWM mode 14
	ICR1 = 25999;
	TCCR1A = _BV(WGM11);
//	start timer1 with prescaler clock/8
	TCCR1B = _BV(WGM13) + _BV(WGM12) + _BV(CS11);

//	activate 16-bit Timer1 overflow interrupt 16 MHz / 8 / 25600 = 78 Hz = 12,8 ms
	TIMSK1 = _BV(TOIE1);
//	i.e. OCR1xH = 100 means 100% of pwm power

//	fast PWM mode 14
	ICR3 = 25599;
	TCCR3A = _BV(WGM31); //_BV(COM3A1) + _BV(COM3B1) + _BV(COM3C1) +
//	start timer3 with prescaler clock/8
	TCCR3B = _BV(WGM33) + _BV(WGM32) + _BV(CS31);

#if slave
//	fast PWM mode 14
	ICR4 = 25599;
	TCCR4A = _BV(WGM41);
	//	start timer4 with prescaler clock/8
	TCCR4B = _BV(WGM43) + _BV(WGM42) + _BV(CS41);
#endif

	GTCCR = _BV(TSM);
	GTCCR = _BV(PSRSYNC); //reset prescaler
	GTCCR &= ~(_BV(TSM)); //synchronous start of all 16-bit timers (0, 1, 3, 4, 5)

	TCNT1 = 0;
	TCNT3 = 0;
	TCNT4 = 0;
	TCNT5 = 0;
	sei();

//	Motorausgnge initial high oder low schalten entsprechend der gewhlten Beschaltung
#if slave
	motorcontrol = output;
#else
	motorcontrol = output1;
#endif
}

/*********************************************************************************
	M () stellt die Funktion zur Manipulation von motorcontrol zur Steuerung 
	einzelner Motoren bereit. Der pwm Wert gilt auch bei Einzelbeschaltung
	der Ausgnge immer fr beide Ausgnge.
*********************************************************************************/
void M (uint8_t motor_number, uint8_t state, uint8_t pwm)
{
//	Abfangen unsinniger Werte
	if (motor_number == 0) return;
	if (motor_number > (4 + (4*slave))) return;
	if (state > 3) return;

	uint8_t position = (motor_number-1)*2;
//	map pwm to OC register
	switch (motor_number)
	{
		case 1: // motor 1
			if (pwm == 0 || pwm == 100) //Deaktiviere die Output Compare Match Interrupts
			{
				TIMSK1 &= ~( _BV(OCIE1A) );
			}
			else
			{
				OCR1AH = pwm;
				OCR1AL = 0;
				TIMSK1 |= _BV(OCIE1A); //aktiviere die Output Compare Match Interrupts
			}
			break;
		case 2: //motor 2
			if (pwm == 0 || pwm == 100) //Deaktiviere die Output Compare Match Interrupts
			{
				TIMSK1 &= ~( _BV(OCIE1B) );
			}
			else
			{
				OCR1BH = pwm;
				OCR1BL = 0;
				TIMSK1 |= _BV(OCIE1B); //aktiviere die Output Compare Match Interrupts
			}
			break;
		case 3: // motor 3
			if (pwm == 0 || pwm == 100) //Deaktiviere die Output Compare Match Interrupts
			{
				TIMSK1 &= ~( _BV(OCIE1C) );
			}
			else
			{
				OCR1CH = pwm;
				OCR1CL = 0;
				TIMSK1 |= _BV(OCIE1C); //aktiviere die Output Compare Match Interrupts
			}
			break;
		case 4: //motor 4
			if (pwm == 0 || pwm == 100) //Deaktiviere die Output Compare Match Interrupts
			{
				TIMSK3 &= ~( _BV(OCIE3A) );
			}
			else
			{
				OCR3AH = pwm;
				OCR3AL = 0;
				TIMSK3 |= _BV(OCIE3A); //aktiviere die Output Compare Match Interrupts
			}
			break;
#if slave
		case 5: //motor 5
			if (pwm == 0 || pwm == 100) //Deaktiviere die Output Compare Match Interrupts
			{
				TIMSK3 &= ~( _BV(OCIE3B) );
			}
			else
			{
				OCR3BH = pwm;
				OCR3BL = 0;
				TIMSK3 |= _BV(OCIE3B); //aktiviere die Output Compare Match Interrupts
			}
			break;

		case 6: //motor 6
			if (pwm == 0 || pwm == 100) //Deaktiviere die Output Compare Match Interrupts
			{
				TIMSK3 &= ~( _BV(OCIE3C) );
			}
			else
			{
				OCR3CH = pwm;
				OCR3CL = 0;
				TIMSK3 |= _BV(OCIE3C); //aktiviere die Output Compare Match Interrupts
			}
			break;
		case 7: // motor 7
			if (pwm == 0 || pwm == 100) //Deaktiviere die Output Compare Match Interrupts
			{
				TIMSK4 &= ~( _BV(OCIE4A) );
			}
			else
			{
				OCR4AH = pwm;
				OCR4AL = 0;
				TIMSK4 |= _BV(OCIE4A); //aktiviere die Output Compare Match Interrupts
			}
			break;
		case 8: //motor 8
			if (pwm == 0 || pwm == 100) //Deaktiviere die Output Compare Match Interrupts
			{
				TIMSK4 &= ~( _BV(OCIE4B) );
			}
			else
			{
				OCR4BH = pwm;
				OCR4BL = 0;
				TIMSK4 |= _BV(OCIE4B); //aktiviere die Output Compare Match Interrupts
			}
#endif
	}

//	Zunchst wird "state" in "motorcontrol" an der gewnschten Position hinzugefuegt
//	dazu werden zunchst die jeweiligen 2 bits an den Bitpositionen fr motor_number
//	auf 0 gesetzt
	motorcontrol &= ~(3 << position);
//	und dann die 2 neuen Bits dem motorcontrol hinzugefgt
	motorcontrol |= state << position;
}

/********************************************************************************
	E () stellt die Funktion zum Dekodieren einzelner Eingnge aus dem inputword
	bereit. Zuvor muss inputword mittels FTUI_update() eingelesen worden sein.
	input_number ist die Eingangsnummer (1-8 bzw 1-16 mit zustzlichem Slave-
	interface am Erweiterungsport) die dekodiert werden soll.
********************************************************************************/
uint8_t E (uint8_t input_number)
{
	uint8_t input;

//	Wenn das inputword an der Bitposition von 
//	"input_number" > 0 dann gib 1 zurck, ansonsten 0
	if (inputword & (1<<(input_number-1))) input = 1;
	else input = 0;

//	Rckgabewert
	return input;
}

